from .neobaseextractor import NeoBaseRecordingExtractor, NeoBaseSortingExtractor

try:
    import neo
    HAVE_NEO = True
except ImportError:
    HAVE_NEO = False


class BlackrockRecordingExtractor(NeoBaseRecordingExtractor):
    """
    The Blackrock extractor is wrapped from neo.rawio.BlackrockRawIO.
    
    Parameters
    ----------
    filename: str
        The Blackrock file (.ns1, .ns2, .ns3, .ns4m .ns4, or .ns6)
    block_index: None or int
        If the underlying dataset have several blocks the index must be specified.
    seg_index_index: None or int
        If the underlying dataset have several segments the index must be specified.
    
    """
    extractor_name = 'BlackrockRecording'
    mode = 'file'
    installed = HAVE_NEO
    NeoRawIOClass = 'BlackrockRawIO'


class BlackrockSortingExtractor(NeoBaseSortingExtractor):
    """
    The Blackrock extractor is wrapped from neo.rawio.BlackrockRawIO.

    Parameters
    ----------
    filename: str
        The Blackrock file (.nev)
    block_index: None or int
        If the underlying dataset have several blocks the index must be specified.
    seg_index_index: None or int
        If the underlying dataset have several segments the index must be specified.

    """
    extractor_name = 'BlackrockSorting'
    mode = 'file'
    installed = HAVE_NEO
    NeoRawIOClass = 'BlackrockRawIO'



from abc import ABC

from spikeextractors import SortingExtractor
from pathlib import Path
import numpy as np

try:
    import pandas as pd

    HAVE_PANDAS = True
except:
    HAVE_PANDAS = False


class ALFSortingExtractor(SortingExtractor):
    extractor_name = 'ALFSortingExtractor'
    installed = HAVE_PANDAS  # check at class level if installed or not
    is_writable = True
    mode = 'folder'
    installation_mesg = "To use the ALFSortingExtractor run:\n\n pip install pandas\n\n"

    def __init__(self, folder_path, sampling_frequency=30000):
        assert HAVE_PANDAS, self.installation_mesg
        SortingExtractor.__init__(self)
        # check correct parent folder:
        self.file_loc = Path(folder_path)
        if 'probe' not in Path(self.file_loc).name:
            raise ValueError('folder name should contain "probe", containing channels, clusters.* .npy datasets')
        # load datasets as mmap into a dict:
        self._required_alf_datasets = ['spikes.times', 'spikes.clusters']
        self._found_alf_datasets = dict()
        for alf_dataset_name in self.file_loc.iterdir():
            if 'spikes' in alf_dataset_name.stem or 'clusters' in alf_dataset_name.stem:
                if 'npy' in alf_dataset_name.suffix:
                    self._found_alf_datasets.update({alf_dataset_name.stem: self._load_npy(alf_dataset_name)})
                elif 'metrics' in alf_dataset_name.stem:
                    self._found_alf_datasets.update({alf_dataset_name.stem: pd.read_csv(alf_dataset_name)})
        # check existence of datasets:
        if not any([i in self._found_alf_datasets for i in self._required_alf_datasets]):
            raise Exception(f'could not find {self._required_alf_datasets} in folder')
        # setting units properties:
        self._total_units = 0
        for alf_dataset_name, alf_dataset in self._found_alf_datasets.items():
            if 'clusters' in alf_dataset_name:
                if 'clusters.metrics' in alf_dataset_name:
                    for property_name, property_values in self._found_alf_datasets[alf_dataset_name].iteritems():
                        self.set_units_property(unit_ids=self.get_unit_ids(),
                                                property_name=property_name,
                                                values=property_values.tolist())
                else:
                    self.set_units_property(unit_ids=self.get_unit_ids(),
                                            property_name=alf_dataset_name.split('.')[1],
                                            values=alf_dataset)
                    if self._total_units == 0:
                        self._total_units = alf_dataset.shape[0]
        self._units_map = {i: j for i, j in zip(self.get_unit_ids(), list(range(self._total_units)))}
        self._units_raster = []
        self._sampling_frequency = sampling_frequency
        self._kwargs = {'folder_path': str(Path(folder_path).absolute()), 'sampling_frequency': sampling_frequency}

    def _load_npy(self, npy_path):
        return np.load(npy_path, mmap_mode='r',allow_pickle=True)

    def _get_clusters_spike_times(self, cluster_idx):
        if len(self._units_raster) == 0:
            spike_cluster_data = self._found_alf_datasets['spikes.clusters']
            spike_times_data = self._found_alf_datasets['spikes.times']
            df = pd.DataFrame({'sp_cluster': spike_cluster_data, 'sp_times': spike_times_data})
            data = df.groupby(['sp_cluster'])['sp_times'].apply(np.array).reset_index(name='sp_times_group')
            self._max_time = 0
            self._units_raster = [None]*self._total_units
            for index, sp_times_list in data.values:
                self._units_raster[index] = sp_times_list
                max_time = max(sp_times_list)
                if max_time > self._max_time:
                    self._max_time = max_time
        return self._units_raster[cluster_idx]

    def get_unit_ids(self):
        if 'clusters.metrics' in self._found_alf_datasets and \
                self._found_alf_datasets['clusters.metrics'].get('cluster_id') is not None:
            unit_ids = self._found_alf_datasets['clusters.metrics'].get('cluster_id').tolist()
        else:
            unit_ids = list(range(self._total_units))
        return unit_ids

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):

        '''Code to extract spike frames from the specified unit.
        It will return spike frames from within three ranges:
            [start_frame, t_start+1, ..., end_frame-1]
            [start_frame, start_frame+1, ..., final_unit_spike_frame - 1]
            [0, 1, ..., end_frame-1]
            [0, 1, ..., final_unit_spike_frame - 1]
        if both start_frame and end_frame are given, if only start_frame is
        given, if only end_frame is given, or if neither start_frame or end_frame
        are given, respectively. Spike frames are returned in the form of an
        array_like of spike frames. In this implementation, start_frame is inclusive
        and end_frame is exclusive conforming to numpy standards.

        '''
        unit_idx = self._units_map.get(unit_id)
        if unit_idx is None:
            raise ValueError(f'enter one of unit_id={self.get_unit_ids()}')
        cluster_sp_times = self._get_clusters_spike_times(unit_idx)
        if cluster_sp_times is None:
            return np.array([])
        max_frame = np.ceil(cluster_sp_times[-1]*self.get_sampling_frequency()).astype('int64')
        min_frame = np.floor(cluster_sp_times[0]*self.get_sampling_frequency()).astype('int64')
        start_frame = min_frame if start_frame is None or start_frame < min_frame else start_frame
        end_frame = max_frame if end_frame is None or end_frame > max_frame else end_frame
        if start_frame > max_frame or end_frame < min_frame:
            raise ValueError(f'Use start_frame to end_frame between {min_frame} and {max_frame}')
        cluster_sp_frames = (cluster_sp_times * self.get_sampling_frequency()).astype('int64')
        frame_idx = np.where((cluster_sp_frames >= start_frame) &
                            (cluster_sp_frames < end_frame))
        return cluster_sp_frames[frame_idx]

    @staticmethod
    def write_sorting(sorting, save_path):
        '''
        This is an example of a function that is not abstract so it is optional if you want to override it. It allows other
        SortingExtractors to use your new SortingExtractor to convert their sorted data into your
        sorting file format.
        '''
        assert HAVE_PANDAS, ALFSortingExtractor.installation_mesg
        # write cluster properties as clusters.<property_name>.npy
        save_path = Path(save_path)
        csv_property_names = ['cluster_id', 'cluster_id.1', 'num_spikes', 'firing_rate',
            'presence_ratio', 'presence_ratio_std', 'frac_isi_viol',
            'contamination_est', 'contamination_est2', 'missed_spikes_est',
            'cum_amp_drift', 'max_amp_drift', 'cum_depth_drift', 'max_depth_drift',
            'ks2_contamination_pct', 'ks2_label','amplitude_cutoff', 'amplitude_std',
            'epoch_name', 'isi_viol']
        clusters_metrics_df = pd.DataFrame()
        for property_name in sorting.get_unit_property_names(0):
            data = sorting.get_units_property(property_name=property_name)
            if property_name not in csv_property_names:
                np.save(save_path/f'clusters.{property_name}', data)
            else:
                clusters_metrics_df[property_name] = data
        clusters_metrics_df.to_csv(save_path/'clusters.metrics.csv')
        # save spikes.times, spikes.clusters
        clusters_number = []
        unit_spike_times = []
        for unit_no, unit_id in enumerate(sorting.get_unit_ids()):
            unit_spike_train = sorting.get_unit_spike_train(unit_id=unit_id)
            if unit_spike_train is not None:
                unit_spike_times.extend(np.array(unit_spike_train)/sorting.get_sampling_frequency())
                clusters_number.extend([unit_no]*len(unit_spike_train))
        unit_spike_train = np.array(unit_spike_times)
        clusters_number = np.array(clusters_number)
        spike_times_ids = np.argsort(unit_spike_train)
        spike_times = unit_spike_train[spike_times_ids]
        spike_clusters = clusters_number[spike_times_ids]
        np.save(save_path/'spikes.times', spike_times)
        np.save(save_path/'spikes.clusters', spike_clusters)



from spikeextractors import RecordingExtractor
from .readSGLX import readMeta, SampRate, makeMemMapRaw, GainCorrectIM, GainCorrectNI, ExtractDigital
import numpy as np
from pathlib import Path
from spikeextractors.extraction_tools import check_get_traces_args, check_get_ttl_args


class SpikeGLXRecordingExtractor(RecordingExtractor):
    extractor_name = 'SpikeGLXRecordingExtractor'
    has_default_locations = True
    installed = True  # check at class level if installed or not
    is_writable = False
    mode = 'file'
    installation_mesg = "To use the SpikeGLXRecordingExtractor run:\n\n pip install mtscomp\n\n"  # error message when not installed

    def __init__(self, file_path: str, dtype: str = 'int16'):
        RecordingExtractor.__init__(self)
        self._npxfile = Path(file_path)
        self._basepath = self._npxfile.parents[0]

        assert dtype in ['int16', 'float'], "'dtype' can be either 'int16' or 'float'"
        self._dtype = dtype
        # Gets file type: 'imec0.ap', 'imec0.lf' or 'nidq'
        assert 'imec0.ap' in self._npxfile.name or  'imec0.lf' in self._npxfile.name or 'nidq' in self._npxfile.name, \
            "'file_path' can be an imec0.ap, imec.lf, or nidq file"
        assert 'bin' in self._npxfile.name, "The 'npx_file should be either the 'ap' or the 'lf' bin file."
        if 'imec0.ap' in str(self._npxfile):
            lfp = False
            ap = True
            self.is_filtered = True
        elif 'imec0.lf' in str(self._npxfile):
            lfp = True
            ap = False
        else:
            lfp = False
            ap = False
        aux = self._npxfile.stem.split('.')[-1]
        if aux == 'nidq':
            self._ftype = aux
        else:
            self._ftype = self._npxfile.stem.split('.')[-2] + '.' + aux

        # Metafile
        self._metafile = self._basepath.joinpath(self._npxfile.stem+'.meta')
        if not self._metafile.exists():
            raise Exception("'meta' file for '"+self._ftype+"' traces should be in the same folder.")
        # Read in metadata, returns a dictionary
        meta = readMeta(self._npxfile)
        self._meta = meta

        # Traces in 16-bit format
        if '.cbin' in self._npxfile.name: # compressed binary format used by IBL
            try:
                import mtscomp
            except:
                raise Exception(self.installation_mesg)
            self._raw = mtscomp.Reader()
            self._raw.open(self._npxfile, self._npxfile.with_suffix('.ch'))
        else:
            self._raw = makeMemMapRaw(self._npxfile, meta)  # [chanList, firstSamp:lastSamp+1]

        # sampling rate and ap channels
        self._sampling_frequency = SampRate(meta)
        tot_chan, ap_chan, lfp_chan, locations = _parse_spikeglx_metafile(self._metafile)
        if ap:
            if ap_chan < tot_chan:
                self._channels = list(range(int(ap_chan)))
                self._timeseries = self._raw[0:ap_chan, :]
            else:
                self._channels = list(range(int(tot_chan)))  # OriginalChans(meta).tolist()
        elif lfp:
            if lfp_chan < tot_chan:
                self._channels = list(range(int(lfp_chan)))
                self._timeseries = self._raw[0:lfp_chan, :]
            else:
                self._channels = list(range(int(tot_chan)))
        else:
            # nidq
            self._channels = list(range(int(tot_chan)))
            self._timeseries = self._raw

        # locations
        if len(locations) > 0:
            self.set_channel_locations(locations)

        # get gains
        if meta['typeThis'] == 'imec':
            gains = GainCorrectIM(self._timeseries, self._channels, meta)
        elif meta['typeThis'] == 'nidq':
            gains = GainCorrectNI(self._timeseries, self._channels, meta)

        # set gains - convert from int16 to uVolt
        self.set_channel_gains(self._channels, gains*1e6)
        self._kwargs = {'file_path': str(Path(file_path).absolute()), 'dtype': dtype}

    def get_channel_ids(self):
        return self._channels

    def get_num_frames(self):
        return self._timeseries.shape[1]

    def get_sampling_frequency(self):
        return self._sampling_frequency

    @check_get_traces_args
    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, dtype=None):
        channel_idxs = np.array([self.get_channel_ids().index(ch) for ch in channel_ids])
        if np.all(channel_ids == self.get_channel_ids()):
            recordings = self._timeseries[:, start_frame:end_frame]
        else:
            if np.all(np.diff(channel_idxs) == 1):
                recordings = self._timeseries[channel_idxs[0]:channel_idxs[0]+len(channel_idxs), start_frame:end_frame]
            else:
                # This block of the execution will return the data as an array, not a memmap
                recordings = self._timeseries[channel_idxs, start_frame:end_frame]
        if dtype is not None:
            assert dtype in ['int16', 'float'], "'dtype' can be either 'int16' or 'float'"
        else:
            dtype = self._dtype
        if dtype == 'int16':
            return recordings
        else:
            gains = np.array(self.get_channel_gains())[channel_idxs]
            return recordings * gains[:, None]

    @check_get_ttl_args
    def get_ttl_events(self, start_frame=None, end_frame=None, channel_id=0):
        channel = [channel_id]
        dw = 0
        dig = ExtractDigital(self._raw, firstSamp=start_frame, lastSamp=end_frame, dwReq=dw, dLineList=channel,
                             meta=self._meta)
        dig = np.squeeze(dig)
        diff_dig = np.diff(dig.astype(int))

        rising = np.where(diff_dig > 0)[0] + start_frame
        falling = np.where(diff_dig < 0)[0] + start_frame

        ttl_frames = np.concatenate((rising, falling))
        ttl_states = np.array([1] * len(rising) + [-1] * len(falling))
        sort_idxs = np.argsort(ttl_frames)
        return ttl_frames[sort_idxs], ttl_states[sort_idxs]


def _parse_spikeglx_metafile(metafile):
    tot_channels = None
    ap_channels = None
    lfp_channels = None
    x_pitch = 21
    y_pitch = 20

    locations = []
    with Path(metafile).open() as f:
        for line in f.readlines():
            if 'nSavedChans' in line:
                tot_channels = int(line.split('=')[-1])
            if 'snsApLfSy' in line:
                ap_channels = int(line.split('=')[-1].split(',')[0].strip())
                lfp_channels = int(line.split(',')[-2].strip())
            if 'imSampRate' in line:
                fs = float(line.split('=')[-1])
            if 'snsShankMap' in line:
                map = line.split('=')[-1]
                chans = map.split(')')[1:]
                for chan in chans:
                    chan = chan[1:]
                    if len(chan) > 0:
                        x_pos = int(chan.split(':')[1])
                        y_pos = int(chan.split(':')[2])
                        locations.append([x_pos*x_pitch, y_pos*y_pitch])
    return tot_channels, ap_channels, lfp_channels, locations



from spikeextractors import SortingExtractor
from pathlib import Path
from spikeextractors.extraction_tools import check_valid_unit_id

try:
    import tridesclous as tdc

    HAVE_TDC = True
except ImportError:
    HAVE_TDC = False


class TridesclousSortingExtractor(SortingExtractor):
    extractor_name = 'TridesclousSortingExtractor'
    installed = HAVE_TDC  # check at class level if installed or not
    is_writable = False
    mode = 'folder'
    installation_mesg = "To use the TridesclousSortingExtractor install tridesclous: \n\n pip install tridesclous\n\n"  # error message when not installed

    def __init__(self, folder_path, chan_grp=None):
        assert HAVE_TDC, self.installation_mesg
        tdc_folder = Path(folder_path)
        SortingExtractor.__init__(self)
        
        dataio = tdc.DataIO(str(tdc_folder))
        if chan_grp is None:
            # if chan_grp is not provided, take the first one if unique
            chan_grps = list(dataio.channel_groups.keys())
            assert len(chan_grps) == 1, 'There are several groups in the folder, specify chan_grp=...'
            chan_grp = chan_grps[0]

        self.chan_grp = chan_grp
        
        catalogue = dataio.load_catalogue(name='initial', chan_grp=chan_grp)
        
        labels = catalogue['clusters']['cluster_label']
        labels = labels[labels >= 0]
        self._unit_ids = list(labels)
        # load all spike in memory (this avoid to lock the folder with memmap throug dataio
        self._all_spikes = dataio.get_spikes(seg_num=0, chan_grp=self.chan_grp, i_start=None, i_stop=None).copy()
    
        self._sampling_frequency = dataio.sample_rate
        self._kwargs = {'folder_path': str(Path(folder_path).absolute()), 'chan_grp': chan_grp}

    def get_unit_ids(self):
        return self._unit_ids

    @check_valid_unit_id
    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        start_frame, end_frame = self._cast_start_end_frame(start_frame, end_frame)
        spikes = self._all_spikes
        spikes = spikes[spikes['cluster_label'] == unit_id]
        spike_times = spikes['index']
        if start_frame is not None:
            spike_times = spike_times[spike_times >= start_frame]
        if end_frame is not None:
            spike_times = spike_times[spike_times < end_frame]
        return spike_times.copy()



from spikeextractors import RecordingExtractor
from spikeextractors.extraction_tools import check_get_traces_args, check_get_ttl_args
import numpy as np
from pathlib import Path
from distutils.version import StrictVersion

try:
    import pyintan
    if StrictVersion(pyintan.__version__) >= '0.2.1':
        HAVE_INTAN = True
    else:
        print("pyintan version requires an update (>=0.2.1). Please upgrade with 'pip install --upgrade pyintan'")
        HAVE_INTAN = False
except ImportError:
    HAVE_INTAN = False


class IntanRecordingExtractor(RecordingExtractor):
    extractor_name = 'IntanRecording'
    has_default_locations = False
    is_writable = False
    mode = 'file'
    installed = HAVE_INTAN  # check at class level if installed or not
    installation_mesg = "To use the Intan extractor, install pyintan: \n\n pip install pyintan\n\n"  # error message when not installed

    def __init__(self, file_path: str, dtype: str = 'float', verbose: bool = False):
        assert HAVE_INTAN, self.installation_mesg
        RecordingExtractor.__init__(self)
        assert Path(file_path).suffix == '.rhs' or Path(file_path).suffix == '.rhd', \
            "Only '.rhd' and '.rhs' files are supported"
        self._recording_file = file_path
        self._recording = pyintan.File(file_path, verbose)
        self._num_frames = len(self._recording.times)
        self._analog_channels = np.array([ch for ch in self._recording._anas_chan if all([other_ch not in ch['name']
                                                                                          for other_ch in
                                                                                          ['ADC', 'VDD', 'AUX']])])
        self._num_channels = len(self._analog_channels)
        self._channel_ids = list(range(self._num_channels))
        self._fs = float(self._recording.sample_rate.rescale('Hz').magnitude)

        assert dtype in ['float', 'uint16'], "'dtype' can be either 'float' or 'uint16'"
        self._dtype = dtype

        if self._dtype == 'uint16':
            for i, ch in enumerate(self._analog_channels):
                self.set_channel_property(i, 'gain', ch['gain'])
                self.set_channel_property(i, 'offset', ch['offset'])

        self._kwargs = {'file_path': str(Path(file_path).absolute()), 'verbose': verbose}

    def get_channel_ids(self):
        return self._channel_ids

    def get_num_frames(self):
        return self._num_frames

    def get_sampling_frequency(self):
        return self._fs

    @check_get_traces_args
    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, dtype=None):
        channel_idxs = np.array([self._channel_ids.index(ch) for ch in channel_ids])
        analog_chans = self._analog_channels[channel_idxs]
        if dtype is None:
            return self._recording._read_analog(channels=analog_chans, i_start=start_frame, i_stop=end_frame,
                                                dtype=self._dtype).T
        else:
            assert dtype in ['float', 'uint16'], "'dtype' can be either 'float' or 'uint16'"
            return self._recording._read_analog(channels=analog_chans, i_start=start_frame, i_stop=end_frame,
                                                dtype=dtype).T

    @check_get_ttl_args
    def get_ttl_events(self, start_frame=None, end_frame=None, channel_id=0):
        channels = [np.unique(ev.channels)[0] for ev in self._recording.digital_in_events]
        assert channel_id in channels, f"Specified 'channel' not found. Available channels are {channels}"
        ev = self._recording.events[channels.index(channel_id)]

        ttl_frames = (ev.times.rescale("s") * self.get_sampling_frequency()).magnitude.astype(int)
        ttl_states = np.sign(ev.channel_states)
        ttl_valid_idxs = np.where((ttl_frames >= start_frame) & (ttl_frames < end_frame))[0]
        return ttl_frames[ttl_valid_idxs], ttl_states[ttl_valid_idxs]



from spikeextractors import RecordingExtractor, SortingExtractor
from pathlib import Path
import numpy as np
from spikeextractors.extraction_tools import check_get_traces_args, check_get_ttl_args, check_valid_unit_id

try:
    import h5py
    HAVE_MAX = True
except ImportError:
    HAVE_MAX = False


class MaxOneRecordingExtractor(RecordingExtractor):
    extractor_name = 'MaxOneRecording'
    has_default_locations = True
    installed = HAVE_MAX  # check at class level if installed or not
    is_writable = False
    mode = 'file'
    installation_mesg = "To use the MaxOneRecordingExtractor install h5py: \n\n pip install h5py\n\n"  # error message when not installed

    def __init__(self, file_path):
        assert HAVE_MAX, self.installation_mesg
        RecordingExtractor.__init__(self)
        self._file_path = file_path
        self._fs = None
        self._positions = None
        self._recordings = None
        self._filehandle = None
        self._mapping = None
        self._initialize()
        self._kwargs = {'file_path': str(Path(file_path).absolute())}

    def __del__(self):
        self._filehandle.close()

    def _initialize(self):
        self._filehandle = h5py.File(self._file_path, 'r')
        self._mapping = self._filehandle['mapping']
        if 'lsb' in self._filehandle['settings'].keys():
            self._lsb = self._filehandle['settings']['lsb'][0] * 1e6
        else:
            print("Couldn't read lsb. Setting lsb to 1")
            self._lsb = 1.
        channels = np.array(self._mapping['channel'])
        electrodes = np.array(self._mapping['electrode'])
        # remove unused channels
        routed_idxs = np.where(electrodes > -1)[0]
        self._channel_ids = list(channels[routed_idxs])
        self._electrode_ids = list(electrodes[routed_idxs])
        self._num_channels = len(self._channel_ids)
        self._fs = float(20000)
        self._signals = self._filehandle['sig']
        self._num_frames = self._signals.shape[1]

        # This happens when only spikes are recorded
        if self._num_frames == 0:
            find_max_frame = True
        else:
            find_max_frame = False

        for i_ch, ch, el in zip(routed_idxs, self._channel_ids, self._electrode_ids):
            self.set_channel_locations([self._mapping['x'][i_ch], self._mapping['y'][i_ch]], ch)
            self.set_channel_property(ch, 'electrode', el)

        if 'proc0' in self._filehandle:
            if 'spikeTimes' in self._filehandle['proc0']:
                spikes = self._filehandle['proc0']['spikeTimes']

                spike_mask = [True] * len(spikes)
                for i, ch in enumerate(spikes['channel']):
                    if ch not in self._channel_ids:
                        spike_mask[i] = False
                spikes_channels = np.array(spikes['channel'])[spike_mask]

                if find_max_frame:
                    self._num_frames = np.ptp(spikes['frameno'])

                # load activity as property
                activity_channels, counts = np.unique(spikes_channels, return_counts=True)
                # transform to spike rate
                duration = float(self._num_frames) / self._fs
                counts = counts.astype(float) / duration
                activity_channels = list(activity_channels)
                for ch in self.get_channel_ids():
                    if ch in activity_channels:
                        self.set_channel_property(ch, 'activity', counts[activity_channels.index(ch)])
                    else:
                        self.set_channel_property(ch, 'activity', 0)

    def get_channel_ids(self):
        return list(self._channel_ids)

    def get_electrode_ids(self):
        return list(self._electrode_ids)

    def get_num_frames(self):
        return self._num_frames

    def get_sampling_frequency(self):
        return self._fs

    @check_get_traces_args
    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None):
        if np.array(channel_ids).size > 1:
            if np.any(np.diff(channel_ids) < 0):
                sorted_channel_ids = np.sort(channel_ids)
                sorted_idx = np.array([list(sorted_channel_ids).index(ch) for ch in channel_ids])
                recordings = (self._signals[sorted_channel_ids, start_frame:end_frame] * self._lsb).astype('float32')
                return recordings[sorted_idx]
            else:
                return (self._signals[np.array(channel_ids), start_frame:end_frame] * self._lsb).astype('float32')
        else:
            return (self._signals[np.array(channel_ids), start_frame:end_frame] * self._lsb).astype('float32')

    @check_get_ttl_args
    def get_ttl_events(self, start_frame=None, end_frame=None, channel_id=0):
        bitvals = self._signals[-2:, 0]
        first_frame = bitvals[1] << 16 | bitvals[0]
        bits = self._filehandle['bits']
        bit_frames = bits['frameno'] - first_frame
        bit_states = bits['bits']
        bit_idxs = np.where((bit_frames >= start_frame) & (bit_frames < end_frame))[0]
        ttl_frames = bit_frames[bit_idxs]
        ttl_states = bit_states['bit_idxs']
        ttl_states[ttl_states == 0] = -1
        return ttl_frames, ttl_states


class MaxOneSortingExtractor(SortingExtractor):
    extractor_name = 'MaxOneSorting'
    installed = HAVE_MAX  # check at class level if installed or not
    is_writable = False
    mode = 'file'
    installation_mesg = "To use the MaxOneSortingExtractor install h5py: \n\n pip install h5py\n\n"  # error message when not installed

    def __init__(self, file_path):
        assert HAVE_MAX, self.installation_mesg
        SortingExtractor.__init__(self)
        self._file_path = file_path
        self._filehandle = None
        self._mapping = None
        self._version = None
        self._initialize()
        self._kwargs = {'file_path': str(Path(file_path).absolute())}

    def _initialize(self):
        self._filehandle = h5py.File(self._file_path, 'r')
        self._mapping = self._filehandle['mapping']
        self._signals = self._filehandle['sig']

        bitvals = self._signals[-2:, 0]
        self._first_frame = bitvals[1] << 16 | bitvals[0]

        channels = np.array(self._mapping['channel'])
        electrodes = np.array(self._mapping['electrode'])
        # remove unused channels
        routed_idxs = np.where(electrodes > -1)[0]
        self._channel_ids = list(channels[routed_idxs])
        self._unit_ids = list(electrodes[routed_idxs])
        self._sampling_frequency = float(20000)

        self._spiketrains = []
        self._unit_ids = []

        try:
            spikes = self._filehandle['proc0']['spikeTimes']
            for u in self._channel_ids:
                spiketrain_idx = np.where(spikes['channel'] == u)[0]
                if len(spiketrain_idx) > 0:
                    self._unit_ids.append(u)
                    spiketrain = spikes['frameno'][spiketrain_idx] - self._first_frame
                    idxs_greater_0 = np.where(spiketrain >= 0)[0]
                    self._spiketrains.append(spiketrain[idxs_greater_0])
                    self.set_unit_spike_features(u, 'amplitude', spikes['amplitude'][spiketrain_idx][idxs_greater_0])
        except:
            raise AttributeError("Spike times information are missing from the .h5 file")

    def get_unit_ids(self):
        return self._unit_ids

    @check_valid_unit_id
    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        start_frame, end_frame = self._cast_start_end_frame(start_frame, end_frame)
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        unit_idx = self._unit_ids.index(unit_id)
        spiketrain = self._spiketrains[unit_idx]
        inds = np.where((start_frame <= spiketrain) & (spiketrain < end_frame))
        return spiketrain[inds]



from spikeextractors import SortingExtractor
import numpy as np
from pathlib import Path
from spikeextractors.extraction_tools import check_valid_unit_id
from typing import Union
from scipy.io import loadmat, savemat

PathType = Union[str, Path]


class CellExplorerSortingExtractor(SortingExtractor):
    """
    Extracts spiking information from .mat files stored in the CellExplorer format.

    Spike times are stored in units of seconds.

    Parameters
    ----------
    spikes_matfile_path : PathType
        Path to the sorting_id.spikes.cellinfo.mat file.
    """

    extractor_name = "CellExplorerSortingExtractor"
    installed = True
    is_writable = True
    mode = "file"
    installation_mesg = ""

    def __init__(self, spikes_matfile_path: PathType):
        SortingExtractor.__init__(self)

        spikes_matfile_path = Path(spikes_matfile_path)
        assert spikes_matfile_path.is_file(), f"The spikes_matfile_path ({spikes_matfile_path}) must exist!"
        folder_path = spikes_matfile_path.parent
        sorting_id = spikes_matfile_path.name.split(".")[0]

        session_info_matfile_path = folder_path / f"{sorting_id}.sessionInfo.mat"
        assert session_info_matfile_path.is_file(), "No sessionInfo.mat file found in the folder!"
        session_info_mat = loadmat(file_name=str(session_info_matfile_path.absolute()))
        assert session_info_mat['sessionInfo']['rates'][0][0]['wideband'], "The sesssionInfo.mat file must contain " \
            "a 'sessionInfo' struct with field 'rates' containing field 'wideband'!"
        self._sampling_frequency = float(
            session_info_mat['sessionInfo']['rates'][0][0]['wideband'][0][0][0][0]
        )  # careful not to confuse it with the lfpsamplingrate; reported in units Hz

        spikes_mat = loadmat(file_name=str(spikes_matfile_path.absolute()))
        assert np.all(np.isin(['UID', 'times'], spikes_mat['spikes'].dtype.names)), \
            "The spikes.cellinfo.mat file must contain a 'spikes' struct with fields 'UID' and 'times'!"

        self._unit_ids = np.asarray(spikes_mat['spikes']['UID'][0][0][0], dtype=int)
        # CellExplorer reports spike times in units seconds; SpikeExtractors uses time units of sampling frames
        # Rounding is necessary to prevent data loss from int-casting floating point errors
        self._spiketrains = [
            (np.array([y[0] for y in x]) * self._sampling_frequency).round().astype(int)
            for x in spikes_mat['spikes']['times'][0][0][0]
         ]

        self._kwargs = dict(spikes_matfile_path=str(spikes_matfile_path.absolute()))

    def get_unit_ids(self):
        return list(self._unit_ids)

    @check_valid_unit_id
    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        start_frame, end_frame = self._cast_start_end_frame(start_frame, end_frame)
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        times = self._spiketrains[self.get_unit_ids().index(unit_id)]
        inds = np.where((start_frame <= times) & (times < end_frame))
        return times[inds]

    @staticmethod
    def write_sorting(sorting: SortingExtractor, save_path: PathType):
        assert save_path.suffixes == ['.spikes', '.cellinfo', '.mat'], \
            "The save_path must correspond to the CellExplorer format of sorting_id.spikes.cellinfo.mat!"

        base_path = save_path.parent
        sorting_id = save_path.name.split(".")[0]
        session_info_save_path = base_path / f"{sorting_id}.sessionInfo.mat"
        spikes_save_path = save_path
        base_path.mkdir(parents=True, exist_ok=True)

        sampling_frequency = sorting.get_sampling_frequency()
        session_info_mat_dict = dict(
            sessionInfo=dict(
                rates=dict(
                    wideband=sampling_frequency
                )
            )
        )
        savemat(file_name=session_info_save_path, mdict=session_info_mat_dict)

        spikes_mat_dict = dict(
            spikes=dict(
                UID=sorting.get_unit_ids(),
                times=[[[y / sampling_frequency] for y in x] for x in sorting.get_units_spike_train()]
            )
        )
        # If, in the future, it is ever desired to allow this to write unit properties, they must conform
        # to the format here: https://cellexplorer.org/datastructure/data-structure-and-format/
        savemat(file_name=spikes_save_path, mdict=spikes_mat_dict)



try:
    from sonpy import lib as sp

    # Data storage and function finder
    DataReadFunctions = {
        sp.DataType.Adc: sp.SonFile.ReadInts,
        sp.DataType.EventFall: sp.SonFile.ReadEvents,
        sp.DataType.EventRise: sp.SonFile.ReadEvents,
        sp.DataType.EventBoth: sp.SonFile.ReadEvents,
        sp.DataType.Marker: sp.SonFile.ReadMarkers,
        sp.DataType.AdcMark: sp.SonFile.ReadWaveMarks,
        sp.DataType.RealMark: sp.SonFile.ReadRealMarks,
        sp.DataType.TextMark: sp.SonFile.ReadTextMarks,
        sp.DataType.RealWave: sp.SonFile.ReadFloats
    }
except:
    pass

# Get the saved time and date
# f.GetTimeDate()


def get_channel_info(f, smrx_ch_ind):
    """
    Extract info from smrx files

    Parameters:
    -----------
    f: str
        SonFile object.
    smrx_ch_ind: int
        Index of smrx channel. Does not match necessarily with extractor id.
    """

    ch_info = {
        'type': f.ChannelType(smrx_ch_ind),           # Get the channel kind
        'ch_number': f.PhysicalChannel(smrx_ch_ind),  # Get the physical channel number associated with this channel
        'title': f.GetChannelTitle(smrx_ch_ind),      # Get the channel title
        'rate': f.GetIdealRate(smrx_ch_ind),          # Get the requested channel rate
        'max_time': f.ChannelMaxTime(smrx_ch_ind),    # Get the time of the last item in the channel
        'divide': f.ChannelDivide(smrx_ch_ind),       # Get the waveform sample interval in file clock ticks
        'time_base': f.GetTimeBase(),                 # Get how many seconds there are per clock tick
        'scale': f.GetChannelScale(smrx_ch_ind),      # Get the channel scale
        'offset': f.GetChannelOffset(smrx_ch_ind),    # Get the channel offset
        'unit': f.GetChannelUnits(smrx_ch_ind),       # Get the channel units
        'y_range': f.GetChannelYRange(smrx_ch_ind),   # Get a suggested Y range for the channel
        'comment': f.GetChannelComment(smrx_ch_ind),  # Get the comment associated with a channel
        'size_bytes:': f.ChannelBytes(smrx_ch_ind),   # Get an estimate of the data bytes stored for the channel
    }

    return ch_info


def get_channel_data(f, smrx_ch_ind, start_frame=0, end_frame=None):
    """
    Extract info from smrx files

    Parameters:
    -----------
    f: str
        SonFile object.
    smrx_ch_ind: int
        Index of smrx channel. Does not match necessarily with extractor id.
    start_frame: int
        The starting frame of the trace to be returned (inclusive).
    end_frame: int
        The ending frame of the trace to be returned (exclusive).
    """

    if end_frame is None:
        end_frame = int(f.ChannelMaxTime(smrx_ch_ind) / f.ChannelDivide(smrx_ch_ind))

    data = DataReadFunctions[f.ChannelType(smrx_ch_ind)](
        self=f,
        chan=smrx_ch_ind,
        nMax=int(f.ChannelMaxTime(smrx_ch_ind) / f.ChannelDivide(smrx_ch_ind)),
        tFrom=int(start_frame * f.ChannelDivide(smrx_ch_ind)),
        tUpto=int(end_frame * f.ChannelDivide(smrx_ch_ind))
    )

    return data



from spikeextractors import RecordingExtractor
from .utils import get_channel_info, get_channel_data
from spikeextractors.extraction_tools import check_get_traces_args

import numpy as np
from pathlib import Path
from typing import Union
from copy import deepcopy

try:
    from sonpy import lib as sp

    HAVE_SONPY = True
except ImportError:
    HAVE_SONPY = False

PathType = Union[str, Path, None]
DtypeType = Union[str, np.dtype, None]


class CEDRecordingExtractor(RecordingExtractor):
    """
    Extracts electrophysiology recordings from .smrx files.
    The recording extractor always returns channel IDs starting from 0.
    The recording data will always be returned in the shape of (num_channels,num_frames).

    Parameters
    ----------
    file_path: str
        Path to the .smrx file to be extracted
    smrx_channel_ids: list of int
        List with indexes of valid smrx channels. Does not match necessarily
        with extractor id.
    """

    extractor_name = 'CEDRecordingExtractor'
    installed = HAVE_SONPY  # check at class level if installed or not
    is_writable = False
    mode = 'file'
    installation_mesg = "To use the CED extractor, install sonpy: \n\n pip install sonpy\n\n"  # error message when not installed

    def __init__(self, file_path: PathType, smrx_channel_ids: list):
        assert HAVE_SONPY, self.installation_mesg
        file_path = Path(file_path)
        assert file_path.is_file() and file_path.suffix == '.smrx', 'file_path must lead to a .smrx file!'
        assert len(smrx_channel_ids) > 0, "'smrx_channel_ids' cannot be an empty list!"

        super().__init__()

        # Open smrx file
        self._recording_file_path = file_path
        self._recording_file = sp.SonFile(sName=str(file_path), bReadOnly=True)
        if self._recording_file.GetOpenError() != 0:
            raise ValueError(f'Error opening file:', sp.GetErrorString(self._recording_file.GetOpenError()))

        # Map Recording channel_id to smrx index / test for invalid indexes /
        # get channel info / set channel gains
        self._channelid_to_smrxind = dict()
        self._channel_smrxinfo = dict()
        self._channel_names = []
        for i, ind in enumerate(smrx_channel_ids):
            if self._recording_file.ChannelType(ind) == sp.DataType.Off:
                raise ValueError(f'Channel {ind} is type Off and cannot be used')
            self._channelid_to_smrxind[i] = ind
            self._channel_smrxinfo[i] = get_channel_info(
                f=self._recording_file,
                smrx_ch_ind=ind
            )
            # Set channel gains: http://ced.co.uk/img/Spike10.pdf
            # from 16-bit encoded int / to ADC +-5V input / to measured Volts
            gains = self._channel_smrxinfo[i]['scale'] / 6553.6
            self.set_channel_gains(
                channel_ids=[i],
                gains=gains
            )
            self._channel_names.append(self._channel_smrxinfo[i]['title'])

        rate0 = self._channel_smrxinfo[0]['rate']
        for chan, info in self._channel_smrxinfo.items():
            assert info['rate'] == rate0, "Inconsistency between 'sampling_frequency' of different channels. The " \
                                          "extractor only supports channels with the same 'rate'"

        self._kwargs = {'file_path': str(Path(file_path).absolute()),
                        'smrx_channel_ids': smrx_channel_ids}

    @property
    def channel_names(self):
        return deepcopy(self._channel_names)

    @check_get_traces_args
    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None):
        '''This function extracts and returns a trace from the recorded data from the
        given channels ids and the given start and end frame. It will return
        traces from within three ranges:

            [start_frame, start_frame+1, ..., end_frame-1]
            [start_frame, start_frame+1, ..., final_recording_frame - 1]
            [0, 1, ..., end_frame-1]
            [0, 1, ..., final_recording_frame - 1]

        if both start_frame and end_frame are given, if only start_frame is
        given, if only end_frame is given, or if neither start_frame or end_frame
        are given, respectively. Traces are returned in a 2D array that
        contains all of the traces from each channel with dimensions
        (num_channels x num_frames). In this implementation, start_frame is inclusive
        and end_frame is exclusive conforming to numpy standards.

        Parameters
        ----------
        start_frame: int
            The starting frame of the trace to be returned (inclusive)
        end_frame: int
            The ending frame of the trace to be returned (exclusive)
        channel_ids: array_like
            A list or 1D array of channel ids (ints) from which each trace will be
            extracted

        Returns
        ----------
        traces: numpy.ndarray
            A 2D array that contains all of the traces from each channel.
            Dimensions are: (num_channels x num_frames)
        '''
        recordings = np.vstack(
            [get_channel_data(
                f=self._recording_file,
                smrx_ch_ind=self._channelid_to_smrxind[i],
                start_frame=start_frame,
                end_frame=end_frame
            ) for i in channel_ids]
        )

        return recordings

    def get_num_frames(self):
        '''This function returns the number of frames in the recording

        Returns
        -------
        num_frames: int
            Number of frames in the recording (duration of recording)
        '''
        return int(self._channel_smrxinfo[0]['max_time'] / self._channel_smrxinfo[0]['divide'])

    def get_sampling_frequency(self):
        '''This function returns the sampling frequency in units of Hz.

        Returns
        -------
        fs: float
            Sampling frequency of the recordings in Hz
        '''
        return self._channel_smrxinfo[0]['rate']

    def get_channel_ids(self):
        '''Returns the list of channel ids. If not specified, the range from 0 to num_channels - 1 is returned.

        Returns
        -------
        channel_ids: list
            Channel list

        '''
        return list(self._channelid_to_smrxind.keys())

    @staticmethod
    def get_all_channels_info(file_path):
        """
        Extract info from all channels in the smrx file. Returns a dictionary with
        valid smrx channel indexes as keys and the respective channel information as
        value.

        Parameters:
        -----------
        f: str
            Path to .smrx file
        """
        f = sp.SonFile(sName=str(file_path), bReadOnly=True)
        n_channels = f.MaxChannels()
        return {
            i: get_channel_info(f, i) for i in range(n_channels)
            if f.ChannelType(i) != sp.DataType.Off
        }



import numpy as np
from pathlib import Path

from spikeextractors import SortingExtractor
from spikeextractors.extractors.numpyextractors import NumpyRecordingExtractor
from spikeextractors.extraction_tools import check_valid_unit_id

try:
    import yaml
    HAVE_YASS = True
except:
    HAVE_YASS = False
    
class YassSortingExtractor(SortingExtractor):

    extractor_name = 'YassExtractor'
    mode = 'folder'
    installed = HAVE_YASS  # check at class level if installed or not

    has_default_locations = False
    is_writable = False
    installation_mesg = "To use the Yass extractor, install pyyaml: \n\n pip install pyyaml\n\n"  # error message when not installed
    
    
    def __init__(self, folder_path):
        SortingExtractor.__init__(self)

        assert HAVE_YASS, self.installation_mesg
        
        self.root_dir = folder_path
        r = Path(self.root_dir)

        self.fname_spike_train = r / 'tmp' / 'output' / 'spike_train.npy'
        self.fname_templates = r /'tmp' / 'output' / 'templates' / 'templates_0sec.npy'
        self.fname_config = r / 'config.yaml'
        
        
        # set defaults to None so they are only loaded if user requires them
        
        self.spike_train = None
        self.temps = None

        # Read CONFIG File
        with open(self.fname_config, 'r') as stream:
            self.config = yaml.safe_load(stream)
        
        self._sampling_frequency = self.config['recordings']['sampling_rate']

    def get_unit_ids(self):

        if self.spike_train is None:
            self.spike_train = np.load(self.fname_spike_train)
        
        unit_ids = np.unique(self.spike_train[:,1])
        
        return unit_ids
    
    def get_temps(self):

        # Electrical images/templates.
        
        if self.temps is None:
            self.temps = np.load(self.fname_templates)
                    
        return self.temps

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):

        '''Code to extract spike frames from the specified unit.
        '''

        if self.spike_train is None:
            self.spike_train = np.load(self.fname_spike_train)
            
        # find unit id spike times
        idx = np.where(self.spike_train[:,1]==unit_id)
        spike_times = self.spike_train[idx,0].squeeze()

        # find spike times
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = 1E50 # use large time
            
        idx2 = np.where(np.logical_and(spike_times>=start_frame, spike_times<end_frame))[0]
        spike_times = spike_times[idx2]
        
        return spike_times
    



from spikeextractors import RecordingExtractor, SortingExtractor
from pathlib import Path
import numpy as np
from spikeextractors.extraction_tools import check_get_traces_args, check_get_ttl_args, check_valid_unit_id

try:
    import h5py
    HAVE_MEA1k = True
except ImportError:
    HAVE_MEA1k = False


class Mea1kRecordingExtractor(RecordingExtractor):
    extractor_name = 'Mea1kRecording'
    has_default_locations = True
    installed = HAVE_MEA1k  # check at class level if installed or not
    is_writable = True
    mode = 'file'
    installation_mesg = "To use the Mea1kRecordingExtractor install h5py: \n\n pip install h5py\n\n"  # error message when not installed

    def __init__(self, file_path, load_spikes=True):
        assert HAVE_MEA1k, self.installation_mesg
        RecordingExtractor.__init__(self)
        self._file_path = file_path
        self._fs = None
        self._positions = None
        self._recordings = None
        self._filehandle = None
        self._mapping = None
        self._signals = None
        self._version = None
        self._timestamps = None
        self._load_spikes = load_spikes
        self._initialize()
        self._kwargs = {'file_path': str(Path(file_path).absolute()),
                        'load_spikes': load_spikes}

    def _initialize(self):
        self._filehandle = h5py.File(self._file_path, mode='r')
        try:
            self._version = self._filehandle['version'][0].decode()
        except:
            try:
                self._version = self._filehandle['chipinformation']['software_version'][0].decode()
            except:
                self._version = '20161003'

        self._lsb = 1
        if int(self._version) == 20160704:
            self._signals = self._filehandle.get('sig')
            settings = self._filehandle.get('settings')

            if 'gain' in settings:
                self._gain = self._filehandle.get('settings/gain')[0]
            else:
                print("Couldn't read gain. Setting gain to 512")
                self._gain = 512

            self._bits = self._filehandle.get('bits', [])

            if 'lsb' in settings:
                self._lsb = self._filehandle['settings']['lsb'][0] * 1e6
            else:
                self._lsb = 3.3 / (1024 * self._gain) * 1e6

            assert 'mapping' in self._filehandle.keys(), "Could not load 'mapping' field"
            self._mapping = self._filehandle['mapping']
            self._fs = 20000
        elif int(self._version) >= 20161003:
            self._mapping = self._filehandle['ephys']['mapping']
            self._fs = float(self._filehandle['ephys']['frame_rate'][()])
            self._signals = self._filehandle['ephys']['signal']
        else:
            raise NotImplementedError(f"Version {self._version} of the Mea1k chip is not supported")

        channels = np.array(self._mapping['channel'])
        electrodes = np.array(self._mapping['electrode'])
        # remove unused channels
        routed_idxs = np.where(electrodes > -1)[0]
        self._channel_ids = list(channels[routed_idxs])
        self._electrode_ids = list(electrodes[routed_idxs])
        self._num_channels = len(self._channel_ids)
        self._num_frames = self._signals.shape[1]

        # This happens when only spikes are recorded
        if self._num_frames == 0:
            find_max_frame = True
        else:
            find_max_frame = False

        for i_ch, ch, el in zip(routed_idxs, self._channel_ids, self._electrode_ids):
            self.set_channel_locations([self._mapping['x'][i_ch], self._mapping['y'][i_ch]], ch)
            self.set_channel_property(ch, 'electrode', el)

        if self._load_spikes:
            if 'proc0' in self._filehandle:
                if 'spikeTimes' in self._filehandle['proc0']:
                    spikes = self._filehandle['proc0']['spikeTimes']

                    spike_mask = [True] * len(spikes)
                    for i, ch in enumerate(spikes['channel']):
                        if ch not in self._channel_ids:
                            spike_mask[i] = False
                    spikes_channels = np.array(spikes['channel'])[spike_mask]

                    if find_max_frame:
                        self._num_frames = np.ptp(spikes['frameno'])

                    # load activity as property
                    activity_channels, counts = np.unique(spikes_channels, return_counts=True)
                    # transform to spike rate
                    duration = float(self._num_frames) / self._fs
                    counts = counts.astype(float) / duration
                    activity_channels = list(activity_channels)
                    for ch in self.get_channel_ids():
                        if ch in activity_channels:
                            self.set_channel_property(ch, 'spike_rate', counts[activity_channels.index(ch)])
                            spike_amplitudes = spikes[np.where(spikes['channel'] == ch)]['amplitude']
                            self.set_channel_property(ch, 'spike_amplitude', np.median(spike_amplitudes))
                        else:
                            self.set_channel_property(ch, 'spike_rate', 0)
                            self.set_channel_property(ch, 'spike_amplitude', 0)

    def correct_for_missing_frames(self, verbose=False):
        """
        Corrects for missing frames. The correct times can be retrieved with the frame_to_time and time_to_frame
        functions.

        Parameters
        ----------
        verbose: bool
            If True, output is verbose
        """
        frame_idxs_span = self._get_frame_number(self.get_num_frames() - 1) - self._get_frame_number(0)
        if frame_idxs_span > self.get_num_frames():
            if verbose:
                print(f"Found missing frames! Correcting for it (this might take a while)")

            framenos = self._get_frame_numbers()
            # find missing frames
            diff_frames = np.diff(framenos)
            missing_frames_idxs = np.where(diff_frames > 1)[0]

            delays_in_frames = []
            for mf_idx in missing_frames_idxs:
                delays_in_frames.append(diff_frames[mf_idx])

            if verbose:
                print(f"Found {len(delays_in_frames)} missing intervals")

            self._timestamps = np.round(np.arange(self.get_num_frames()) / self.get_sampling_frequency(), 6)

            for mf_idx, duration in zip(missing_frames_idxs, delays_in_frames):
                self._timestamps[mf_idx:] += np.round(duration / self.get_sampling_frequency(), 6)
        else:
            if verbose:
                print("No missing frames found")

    def frame_to_time(self, frames):
        if self._timestamps is None:
            return super().frame_to_time(frames)
        else:
            return self._timestamps[frames]

    def time_to_frame(self, times):
        if self._timestamps is None:
            return super().time_to_frame(times)
        else:
            return np.searchsorted(self._timestamps, times).astype('int64')

    def get_channel_ids(self):
        return list(self._channel_ids)

    def get_electrode_ids(self):
        return list(self._electrode_ids)

    def get_num_frames(self):
        return self._num_frames

    def get_sampling_frequency(self):
        return self._fs

    @check_get_traces_args
    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None):
        if np.array(channel_ids).size > 1:
            if np.any(np.diff(channel_ids) < 0):
                sorted_channel_ids = np.sort(channel_ids)
                sorted_idx = np.array([list(sorted_channel_ids).index(ch) for ch in channel_ids])
                recordings = (self._signals[sorted_channel_ids, start_frame:end_frame] * self._lsb).astype('float32')
                return recordings[sorted_idx]
            else:
                return (self._signals[np.array(channel_ids), start_frame:end_frame] * self._lsb).astype('float32')
        else:
            return (self._signals[np.array(channel_ids), start_frame:end_frame] * self._lsb).astype('float32')

    @check_get_ttl_args
    def get_ttl_events(self, start_frame=None, end_frame=None, channel_id=0):
        bitvals = self._signals[-2:, 0]
        first_frame = bitvals[1] << 16 | bitvals[0]
        bits = self._filehandle['bits']
        bit_frames = bits['frameno'] - first_frame
        bit_states = bits['bits']
        bit_idxs = np.where((bit_frames >= start_frame) & (bit_frames < end_frame))[0]
        ttl_frames = bit_frames[bit_idxs]
        ttl_states = bit_states['bit_idxs']
        ttl_states[ttl_states == 0] = -1
        return ttl_frames, ttl_states

    def _get_frame_numbers(self):
        bitvals = self._signals[-2:, :]
        frame_nos = np.bitwise_or(np.left_shift(bitvals[-1].astype('int64'), 16), bitvals[0])
        return frame_nos

    def _get_frame_number(self, index):
        bitvals = self._signals[-2:, index]
        frameno = bitvals[1] << 16 | bitvals[0]
        return frameno

    @staticmethod
    def write_recording(recording, save_path, chunk_size=None, chunk_mb=500):
        assert HAVE_MEA1k, Mea1kRecordingExtractor.installation_mesg
        save_path = Path(save_path)
        if save_path.suffix == '':
            save_path = Path(str(save_path) + '.h5')
        mapping_dtype = np.dtype([('electrode', np.int32), ('x', np.float64), ('y', np.float64),
                                  ('channel', np.int32)])

        assert 'location' in recording.get_shared_channel_property_names(), "'location' property is needed to write " \
                                                                            "max1k format"

        with h5py.File(save_path, 'w') as f:
            f.create_group('ephys')
            f.create_dataset('version', data=str(20161003))
            ephys = f['ephys']
            ephys.create_dataset('frame_rate', data=recording.get_sampling_frequency())
            ephys.create_dataset('frame_numbers', data=np.arange(recording.get_num_frames()))
            # save mapping
            mapping = np.empty(recording.get_num_channels(), dtype=mapping_dtype)
            x = recording.get_channel_locations()[:, 0]
            y = recording.get_channel_locations()[:, 1]
            for i, ch in enumerate(recording.get_channel_ids()):
                mapping[i] = (ch, x[i], y[i], ch)
            ephys.create_dataset('mapping', data=mapping)
            # save traces
            recording.write_to_h5_dataset_format('/ephys/signal', file_handle=f, time_axis=1,
                                                 chunk_size=chunk_size, chunk_mb=chunk_mb)


class Mea1kSortingExtractor(SortingExtractor):
    extractor_name = 'Mea1kSorting'
    installed = HAVE_MEA1k  # check at class level if installed or not
    is_writable = False
    mode = 'file'
    installation_mesg = "To use the Mea1kSortingExtractor install h5py: \n\n pip install h5py\n\n"  # error message when not installed

    def __init__(self, file_path):
        assert HAVE_MEA1k, self.installation_mesg
        SortingExtractor.__init__(self)
        self._file_path = file_path
        self._filehandle = None
        self._mapping = None
        self._version = None
        self._initialize()
        self._kwargs = {'file_path': str(Path(file_path).absolute())}

    def _initialize(self):
        self._filehandle = h5py.File(self._file_path, mode='r')
        try:
            self._version = self._filehandle['version'][0].decode()
        except:
            try:
                self._version = self._filehandle['chipinformation']['software_version'][0].decode()
            except:
                self._version = '20161003'

        print(f"Chip version: {self._version}")
        self._lsb = 1
        if int(self._version) == 20160704:
            assert 'mapping' in self._filehandle.keys(), "Could not load 'mapping' field"
            self._mapping = self._filehandle['mapping']
            self._signals = self._filehandle.get('sig')
            self._sampling_frequency = 20000
        elif int(self._version) >= 20161003:
            self._mapping = self._filehandle['ephys']['mapping']
            self._sampling_frequency = float(self._filehandle['ephys']['frame_rate'][()])
            self._signals = self._filehandle['ephys']['signal']
        else:
            raise NotImplementedError(f"Version {self._version} of the Mea1k chip is not supported")

        self._first_frame = 0
        try:
            bitvals = self._signals[-2:, 0]
            self._first_frame = bitvals[1] << 16 | bitvals[0]
        except:
            print("Couldn't find first frame information. Setting to 0.")
        channels = np.array(self._mapping['channel'])
        electrodes = np.array(self._mapping['electrode'])
        # remove unused channels
        routed_idxs = np.where(electrodes > -1)[0]
        self._channel_ids = list(channels[routed_idxs])
        self._electrode_ids = list(electrodes[routed_idxs])

        self._spiketrains = []
        self._unit_ids = []

        try:
            spikes = self._filehandle['proc0']['spikeTimes']
            for u in self._channel_ids:
                spiketrain_idx = np.where(spikes['channel'] == u)[0]
                if len(spiketrain_idx) > 0:
                    self._unit_ids.append(u)
                    spiketrain = spikes['frameno'][spiketrain_idx] - self._first_frame
                    idxs_greater_0 = np.where(spiketrain >= 0)[0]
                    self._spiketrains.append(spiketrain[idxs_greater_0])
                    self.set_unit_spike_features(u, 'amplitude', spikes['amplitude'][spiketrain_idx][idxs_greater_0])
        except:
            raise AttributeError("Spike times information are missing from the .h5 file")

    def get_unit_ids(self):
        return self._unit_ids

    @check_valid_unit_id
    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        start_frame, end_frame = self._cast_start_end_frame(start_frame, end_frame)
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        unit_idx = self._unit_ids.index(unit_id)
        spiketrain = self._spiketrains[unit_idx]
        inds = np.where((start_frame <= spiketrain) & (spiketrain < end_frame))
        return spiketrain[inds]



from spikeextractors.extractors.bindatrecordingextractor import BinDatRecordingExtractor
from spikeextractors.extractors.npzsortingextractor import NpzSortingExtractor
from spikeextractors import RecordingExtractor, SortingExtractor
import tempfile
from pathlib import Path
from copy import deepcopy
import importlib
import shutil


class CacheRecordingExtractor(BinDatRecordingExtractor, RecordingExtractor):
    def __init__(self, recording, chunk_size=None, chunk_mb=500, save_path=None, n_jobs=1, joblib_backend='loky',
                 verbose=False):
        RecordingExtractor.__init__(self)  # init tmp folder before constructing BinDatRecordingExtractor
        tmp_folder = self.get_tmp_folder()
        self._recording = recording
        if save_path is None:
            self._is_tmp = True
            self._tmp_file = tempfile.NamedTemporaryFile(suffix=".dat", dir=tmp_folder).name
        else:
            save_path = Path(save_path)
            if save_path.suffix != '.dat' and save_path.suffix != '.bin':
                save_path = save_path.with_suffix('.dat')
            save_path.parent.mkdir(parents=True, exist_ok=True)
            self._is_tmp = False
            self._tmp_file = save_path
        self._dtype = recording.get_dtype()
        recording.write_to_binary_dat_format(save_path=self._tmp_file, dtype=self._dtype, chunk_size=chunk_size,
                                             chunk_mb=chunk_mb, n_jobs=n_jobs, joblib_backend=joblib_backend,
                                             verbose=verbose)
        # keep track of filter status when dumping
        self.is_filtered = self._recording.is_filtered
        BinDatRecordingExtractor.__init__(self, self._tmp_file, numchan=recording.get_num_channels(),
                                          recording_channels=recording.get_channel_ids(),
                                          sampling_frequency=recording.get_sampling_frequency(),
                                          dtype=self._dtype, is_filtered=self.is_filtered)
        # keep BinDatRecording kwargs
        self._bindat_kwargs = deepcopy(self._kwargs)
        self.set_tmp_folder(tmp_folder)
        self.copy_channel_properties(recording)
        self._kwargs = {'recording': recording, 'chunk_size': chunk_size, 'chunk_mb': chunk_mb}

    def __del__(self):
        if self._is_tmp:
            try:
                # close memmap file (for Windows)
                del self._timeseries
                Path(self._tmp_file).unlink()
            except Exception as e:
                print("Unable to remove temporary file", e)

    @property
    def filename(self):
        return str(self._tmp_file)

    def move_to(self, save_path):
        save_path = Path(save_path)
        if save_path.suffix != '.dat' and save_path.suffix != '.bin':
            save_path = save_path.with_suffix('.dat')
        save_path.parent.mkdir(parents=True, exist_ok=True)
        # close memmap file (for Windows)
        del self._timeseries
        shutil.move(self._tmp_file, str(save_path))
        self._tmp_file = str(save_path)
        self._kwargs['file_path'] = str(Path(self._tmp_file).absolute())
        self._bindat_kwargs['file_path'] = str(Path(self._tmp_file).absolute())
        self._is_tmp = False
        tmp_folder = self.get_tmp_folder()
        # re-initialize with new file
        BinDatRecordingExtractor.__init__(self, self._tmp_file, numchan=self._recording.get_num_channels(),
                                          recording_channels=self._recording.get_channel_ids(),
                                          sampling_frequency=self._recording.get_sampling_frequency(),
                                          dtype=self._dtype, is_filtered=self.is_filtered)
        self.set_tmp_folder(tmp_folder)
        self.copy_channel_properties(self._recording)

    # override to make serialization avoid reloading and saving binary file
    def make_serialized_dict(self, include_properties=None, include_features=None):
        '''
        Makes a nested serialized dictionary out of the extractor. The dictionary be used to re-initialize an
        extractor with spikeextractors.load_extractor_from_dict(dump_dict)

        Returns
        -------
        dump_dict: dict
            Serialized dictionary
        include_properties: list or None
            List of properties to include in the dictionary
        include_features: list or None
            List of features to include in the dictionary
        '''
        class_name = str(BinDatRecordingExtractor).replace("<class '", "").replace("'>", '')
        module = class_name.split('.')[0]
        imported_module = importlib.import_module(module)

        if self._is_tmp:
            print("Warning: dumping a CacheRecordingExtractor. The path to the tmp binary file will be lost in "
                  "further sessions. To prevent this, use the 'CacheRecordingExtractor.move_to('path-to-file)' "
                  "function")

        dump_dict = {'class': class_name, 'module': module, 'kwargs': self._bindat_kwargs,
                     'key_properties': self._key_properties, 'version': imported_module.__version__, 'dumpable': True}
        return dump_dict


class CacheSortingExtractor(NpzSortingExtractor, SortingExtractor):
    def __init__(self, sorting, save_path=None):
        SortingExtractor.__init__(self)  # init tmp folder before constructing NpzSortingExtractor
        tmp_folder = self.get_tmp_folder()
        self._sorting = sorting
        if save_path is None:
            self._is_tmp = True
            self._tmp_file = tempfile.NamedTemporaryFile(suffix=".npz", dir=tmp_folder).name
        else:
            save_path = Path(save_path)
            if save_path.suffix != '.npz':
                save_path = save_path.with_suffix('.npz')
            save_path.parent.mkdir(parents=True, exist_ok=True)
            self._is_tmp = False
            self._tmp_file = save_path
        NpzSortingExtractor.write_sorting(self._sorting, self._tmp_file)
        NpzSortingExtractor.__init__(self, self._tmp_file)
        # keep Npz kwargs
        self._npz_kwargs = deepcopy(self._kwargs)
        self.set_tmp_folder(tmp_folder)
        self.copy_unit_properties(sorting)
        self.copy_unit_spike_features(sorting)
        self._kwargs = {'sorting': sorting}

    def __del__(self):
        if self._is_tmp:
            try:
                Path(self._tmp_file).unlink()
            except Exception as e:
                print("Unable to remove temporary file", e)

    @property
    def filename(self):
        return str(self._tmp_file)

    def move_to(self, save_path):
        save_path = Path(save_path)
        if save_path.suffix != '.npz':
            save_path = save_path.with_suffix('.npz')
        save_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.move(self._tmp_file, str(save_path))
        self._tmp_file = str(save_path)
        self._kwargs['file_path'] = str(Path(self._tmp_file).absolute())
        self._npz_kwargs['file_path'] = str(Path(self._tmp_file).absolute())
        self._is_tmp = False
        tmp_folder = self.get_tmp_folder()
        # re-initialize with new file
        NpzSortingExtractor.__init__(self, self._tmp_file)
        # keep Npz kwargs
        self.set_tmp_folder(tmp_folder)
        self.copy_unit_properties(self._sorting)
        self.copy_unit_spike_features(self._sorting)

    # override to make serialization avoid reloading and saving npz file
    def make_serialized_dict(self, include_properties=None, include_features=None):
        '''
        Makes a nested serialized dictionary out of the extractor. The dictionary be used to re-initialize an
        extractor with spikeextractors.load_extractor_from_dict(dump_dict)

        Returns
        -------
        dump_dict: dict
            Serialized dictionary
        include_properties: list or None
            List of properties to include in the dictionary
        include_features: list or None
            List of features to include in the dictionary
        '''
        class_name = str(NpzSortingExtractor).replace("<class '", "").replace("'>", '')
        module = class_name.split('.')[0]
        imported_module = importlib.import_module(module)

        if self._is_tmp:
            print("Warning: dumping a CacheSortingExtractor. The path to the tmp binary file will be lost in "
                  "further sessions. To prevent this, use the 'CacheSortingExtractor.move_to('path-to-file)' "
                  "function")

        dump_dict = {'class': class_name, 'module': module, 'kwargs': self._npz_kwargs,
                     'key_properties': self._key_properties, 'version': imported_module.__version__, 'dumpable': True}
        return dump_dict



from .recordingextractor import RecordingExtractor
from .sortingextractor import SortingExtractor
from .cacheextractors import CacheRecordingExtractor, CacheSortingExtractor
from .subsortingextractor import SubSortingExtractor
from .subrecordingextractor import SubRecordingExtractor
from .multirecordingchannelextractor import concatenate_recordings_by_channel, MultiRecordingChannelExtractor
from .multirecordingtimeextractor import concatenate_recordings_by_time, MultiRecordingTimeExtractor
from .multisortingextractor import concatenate_sortings, MultiSortingExtractor

from .extractorlist import *

from . import example_datasets
from .extraction_tools import load_probe_file, save_to_probe_file, read_binary, write_to_binary_dat_format,\
    write_to_h5_dataset_format, get_sub_extractors_by_property, load_extractor_from_json, load_extractor_from_dict, \
    load_extractor_from_pickle

from distutils.version import StrictVersion

try:
    import h5py
    if StrictVersion(h5py.__version__) > '2.10.0':
        print("h5py version > 2.10.0. Some extractors might not work properly. It is recommended to "
              "downgrade to version 2.10.0: \n>>> pip install h5py==2.10.0")
except ImportError:
    pass

from .version import version as __version__



from pathlib import Path
import numpy as np
from spikeextractors import SortingExtractor
from spikeextractors.extraction_tools import check_valid_unit_id
from typing import Union


try:
    import h5py
    HAVE_H5PY = True
except ImportError:
    HAVE_H5PY = False

PathType = Union[str, Path]


class CombinatoSortingExtractor(SortingExtractor):
    extractor_name = 'CombinatoSortingExtractor'
    installation_mesg = ""  # error message when not installed
    installed = HAVE_H5PY
    is_writable = False
    installation_mesg = "To use the CombinatoSortingExtractor install h5py: \n\n pip install h5py\n\n"  # error message when not installed

    def __init__(self, datapath: PathType, sampling_frequency=None, user='simple',det_sign = 'both'):
        super().__init__()
        datapath = Path(datapath)
        assert datapath.is_dir(), 'Folder {} doesn\'t exist'.format(datapath)
        if sampling_frequency is None:
            h5_path = str(datapath) + '.h5'
            if Path(h5_path).exists():
                with h5py.File(h5_path, mode='r') as f:
                    sampling_frequency = f['sr'][0]
        self.set_sampling_frequency(sampling_frequency)
        det_file = str(datapath / Path('data_' + datapath.stem + '.h5'))
        sort_cat_files = []
        for sign in ['neg', 'pos']:
            if det_sign in ['both', sign]:
                sort_cat_file = datapath / Path('sort_{}_{}/sort_cat.h5'.format(sign,user))
                if sort_cat_file.exists():
                    sort_cat_files.append((sign, str(sort_cat_file)))
        unit_counter = 0
        self._spike_trains = {}
        metadata = {}
        unsorted = []
        fdet = h5py.File(det_file, mode='r')
        for sign, sfile in sort_cat_files:
            with h5py.File(sfile, mode='r') as f:
                sp_class = f['classes'][()]
                gaux = f['groups'][()]
                groups = {g:gaux[gaux[:, 1] == g, 0] for g in np.unique(gaux[:, 1])} #array of classes per group
                group_type = {group: g_type for group,g_type in f['types'][()]}
                sp_index = f['index'][()]

            times_css = fdet[sign]['times'][()]
            for gr, cls in groups.items():
                if group_type[gr] == -1: #artifacts
                    continue
                elif group_type[gr] == 0: #unsorted
                    unsorted.append(np.rint(times_css[sp_index[np.isin(sp_class,cls)]] * (sampling_frequency/1000)))
                    continue

                unit_counter = unit_counter + 1
                self._spike_trains[unit_counter] = np.rint(times_css[sp_index[np.isin(sp_class, cls)]] * (sampling_frequency / 1000))
                metadata[unit_counter] = {'det_sign': sign,
                                          'group_type': 'single-unit' if group_type[gr] else 'multi-unit'}

        fdet.close()

        self._unsorted_train = np.array([])
        if len(unsorted) == 1:
            self._unsorted_train = unsorted[0]
        elif len(unsorted) == 2: #unsorted in both signs
            self._unsorted_train = np.sort(np.concatenate(unsorted), kind='mergesort')

        self._unit_ids = list(range(1, unit_counter+1))
        for u in self._unit_ids:
            for prop,value in metadata[u].items():
                self.set_unit_property(u, prop, value)

    def get_unit_ids(self):
        return self._unit_ids


    @check_valid_unit_id
    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        start_frame, end_frame = self._cast_start_end_frame(start_frame, end_frame)

        start_frame = start_frame or 0
        end_frame = end_frame or np.infty
        st = self._spike_trains[unit_id]
        return st[(st >= start_frame) & (st < end_frame)]

    def get_unsorted_spike_train(self, start_frame=None, end_frame=None):
        start_frame, end_frame = self._cast_start_end_frame(start_frame, end_frame)

        start_frame = start_frame or 0
        end_frame = end_frame or np.infty
        u = self._unsorted_train
        return u[(u >= start_frame) & (u < end_frame)]



